import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestStreamerSlice:
    # start
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_skip_empty_start(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(start=2)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_skip_hasSingleItem_start(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(start=2)) == []

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_skip_hasMultipleItems_start(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(start=2)) == iterable_list[2:]

    # stop
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_skip_empty_stop(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(stop=2)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_skip_hasSingleItem_stop(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(stop=1)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_skip_hasMultipleItems_stop(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(stop=2)) == iterable_list[:2]

    # step
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_skip_empty_step(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(step=2)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_skip_hasSingleItem_step(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(step=1)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_skip_hasMultipleItems_step(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).slice(step=2)) == iterable_list[::2]

    # start, stop, step
    def test_skip_hasMultipleItems_startStopStep(self):
        assert list(q(range(10)).slice(start=1, stop=6, step=2)) == [1, 3, 5]


from copy import copy
from pathlib import Path

from pydoc_markdown.interfaces import Context
from pydoc_markdown.contrib.loaders.python import PythonLoader
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer

context = Context(directory='.')
# loader = PythonLoader(search_path=['../fliq'])  # for debugging
loader = PythonLoader(search_path=['fliq'])
renderer = MarkdownRenderer(render_module_header=False,
                            descriptive_class_title=False,
                            classdef_code_block=False,
                            # signature_with_def=False, # Lib bug here, can't disable
                            render_toc=True,
                            render_toc_title=False,
                            header_level_by_type={
                                "Module": 3,
                                "Class": 3,
                                "Method": 3,
                                "Function": 3,
                                "Variable": 3
                            }
                            )

loader.init(context)
renderer.init(context)

modules = loader.load()
query_module = [m for m in modules if m.name == 'query'][0]
query_class = [c for c in query_module.members if c.name == 'Query'][0]

query_carrier = copy(query_class)
query_carrier.name = 'Streamers'
query_carrier.members = [m for m in query_class.members
                         if not m.name.startswith('_') and m.return_type == "'Query'"]

query_collector = copy(query_class)
query_collector.name = 'Collectors'
query_collector.members = [m for m in query_class.members
                           if not m.name.startswith('_') and m.return_type != "'Query'"]

current_dir = Path(__file__).parent
root_dir = current_dir.parent
api_md_path = root_dir / "api.md"
readme_base_md_path = root_dir / "readme_base.md"
readme_md_path = root_dir / "README.md"

api_content = renderer.render_to_string([query_carrier, query_collector])
with open(api_md_path, 'w') as f:
    f.writelines(api_content)


# merge api.md and readme_base.md to README.md,
# by replace {{autoapi}} in readme_base.md with api_content
def replace_autoapi_in_readme(content: str) -> str:
    with open(readme_base_md_path, 'r') as readme_base:
        readme_template = readme_base.read()
        readme_content = readme_template.replace('{{auto_api}}', content)

    with open(readme_md_path, 'w') as readme:
        readme.writelines(readme_content)


replace_autoapi_in_readme(api_content)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestStreamerTake:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_take_hasNoItems(self,
                             iter_type,
                             iterable,
                             iterable_list):
        assert list(q(iterable).take(n=5)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_take_hasSingleItem(self,
                                iter_type,
                                iterable,
                                iterable_list):
        assert list(q(iterable).take(n=5)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_take_hasMultipleItems_withoutPredicate(self,
                                                    iter_type,
                                                    iterable,
                                                    iterable_list):
        query = q(iterable).take(n=3)
        assert list(query) == iterable_list[:3]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_take_hasMultipleItems_withPredicate(self,
                                                 iter_type,
                                                 iterable,
                                                 iterable_list):
        query = q(iterable).take(n=1, predicate=lambda x: int(x) > 1)
        assert list(query) == iterable_list[2:3]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_take_hasMultipleItems_notEnoughItems(self,
                                                  iter_type,
                                                  iterable,
                                                  iterable_list):
        assert list(q(iterable).take(n=100)) == iterable_list



import pytest

from fliq.tests.fliq_test_utils import Params
from fliq import q


class TestStreamerWhere:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_where_hasNoItems_withPredicate(self,
                                            iter_type,
                                            iterable,
                                            iterable_list):
        assert list(q(iterable).where(lambda x: int(x) > 0)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_where_hasNoItems_withoutPredicate(self,
                                               iter_type,
                                               iterable,
                                               iterable_list):
        assert list(q(iterable).where()) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_where_hasSingleItem_withoutPredicate(self,
                                                  iter_type,
                                                  iterable,
                                                  iterable_list):
        assert list(q(iterable).where()) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_where_hasSingleItem_withPredicate(self,
                                               iter_type,
                                               iterable,
                                               iterable_list):
        assert list(q(iterable).where(lambda x: int(x) < 0)) == []

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_where_hasMultipleItems_withoutFilter(self,
                                                  iter_type,
                                                  iterable,
                                                  iterable_list):
        query = q(iterable).where(lambda x: int(x) > 2)
        assert list(query) == iterable_list[3:]



from typing import Iterable


class QueryBase:
    def __init__(self, iterable: Iterable):
        if iterable is None:
            raise TypeError("iterable cannot be None")

        if not isinstance(iterable, Iterable):
            raise TypeError("iterable must be an iterable")

        self._items = iterable



from dataclasses import dataclass
from fliq import q
import timeit


@dataclass
class Person:
    name: str
    age: int


def gen_data(num: int):
    return [Person(f"Person {i}", i) for i in range(num)]


# get the last letter of the name of the first person over 25 years old


def using_fliq(num: int):
    return lambda: (
        q(gen_data(num))
        .where(lambda p: p.age > 25)
        .select(lambda p: p.name)
        .select(lambda name: name[-1])
        .first_or_default(default=Person("No one", 0))
    )


# Standard library
def using_standard_lib(num: int):
    return lambda: next(
        map(lambda name: name[-1],
            map(lambda p: p.name,
                filter(lambda p: p.age > 25, gen_data(num))))) or Person("No one", 0)


results = []
for b in [100, 1000, 10_000, 100_000, 1_000_000]:
    print(f"running benchmark: {b} items")
    fliq_time = timeit.timeit(using_fliq(b), number=100)
    std_lib_time = timeit.timeit(using_standard_lib(b), number=100)
    print(f"Using fliq: {fliq_time:.5f} seconds")
    print(f"Using standard lib: {std_lib_time:.5f} seconds")
    print(f"Ratio: fliq is {fliq_time / std_lib_time:.2f} times slower")
    results.append(fliq_time / std_lib_time)
    print()

print(f"on avg, fliq is {sum(results) / len(results):.2f} "
      f"times slower than standard lib")



import pytest

from fliq.tests.fliq_test_utils import Params
from fliq import q


class TestStreamerExclude:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_exclude_hasNoItems(self,
                                iter_type,
                                iterable,
                                iterable_list):
        assert list(q(iterable).exclude(lambda x: int(x) > 0)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_exclude_hasSingleItem(self,
                                   iter_type,
                                   iterable,
                                   iterable_list):
        assert list(q(iterable).exclude(lambda x: int(x) < 1)) == iterable_list[1:]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_exclude_hasMultipleItem(self,
                                     iter_type,
                                     iterable,
                                     iterable_list):
        assert list(q(iterable).exclude(lambda x: int(x) < 1)) == iterable_list[1:]



import collections.abc
import typing
from itertools import islice
from typing import Iterable, List, Optional, Any, Sized, Iterator, Callable

from fliq.exceptions import NoItemsFoundException, MultipleItemsFoundException
from fliq.types import Predicate

if typing.TYPE_CHECKING:
    from fliq import q  # noqa: F401 (used in docs)  # pragma: no cover


class Query(collections.abc.Iterable):
    def __init__(self, iterable: Iterable):
        """
        Create a Query object to allow fluent iterable processing
        """
        self._items = iterable
        self._iterator: Optional[Iterator] = None

    def __iter__(self):
        if self._iterator is None:
            self._iterator = iter(self._items)
        return self

    def __next__(self):
        if self._iterator is None:
            self._iterator = iter(self._items)
        return next(self._iterator)

    # region Streamers

    def where(self, predicate: Optional[Predicate] = None) -> 'Query':
        """
        Yields elements that satisfy the predicate (aka filter).

        Example:
            <br />
            `q(range(10)).where(lambda x: x % 2 == 0)`
            <br />
            `[0, 2, 4, 6, 8]`

        Args:
            <br />
            predicate: Optional. The predicate to filter the iterable by. If None is
            given, no filtering takes place.
        """
        if predicate is None:
            # supported to ease syntax in higher level streamers and collectors
            return self

        self._items = filter(predicate, self._items)
        return self

    def select(self, selector: Callable[[Any], Any]) -> 'Query':
        """
        Yields the result of applying the selector function to each element (aka map).

        Example:
            <br />
            `q(range(5)).select(lambda x: x * 2 == 0)`
            <br />
            `[0, 2, 4, 6, 8, 10]`

        Args:
            <br />
            selector: The selector function to apply to each element.
        """
        self._items = map(selector, self._items)
        return self

    def exclude(self, predicate: Predicate) -> 'Query':
        """
        Yields elements that do not satisfy the predicate.

        Example:
            <br />
            `q(range(5)).exclude(lambda x: x > 3)`
            <br />
            `[0, 1, 2, 3]`

        Args:
            <br />
            predicate: The predicate to filter the iterable by.
        """
        self._items = filter(lambda x: not predicate(x), self._items)
        return self

    def distinct(self, preserve_order: bool = True) -> 'Query':
        """
        Yields distinct elements, preserving order if specified.

        Example:
            <br />
            `q([0, 1, 0, 2, 2]).distinct()`
            <br />
            `[0, 1, 2]`

        Args:
            <br />
            preserve_order: Optional. Whether to preserve the order of the items. Defaults to True.
        """
        self._items = dict.fromkeys(self._items).keys() if preserve_order else set(self._items)
        return self

    def order(self,
              by: Optional[Callable[[Any], Any]] = None,
              ascending: bool = True) -> 'Query':
        """Yields elements in sorted order.

        Example:
            <br />
            `q([4, 3, 2, 1, 0]).order()`
            <br />
            `[0, 1, 2, 3, 4]`

        Args:
            <br />
            by: a selector function to extract the key from an item, defaults to None.
            If None, the default ordering is used.
            <br />
            ascending: whether to sort in ascending or descending order, defaults to True.
        """
        if by is None:
            # natural order
            self._items = sorted(self._items, reverse=not ascending)
        else:
            self._items = sorted(self._items, key=by, reverse=not ascending)
        return self

    def reverse(self) -> 'Query':
        """
        Yields elements in reverse order.
        Notes:
         - in case of an irreversible iterable, TypeError is raised (e.g., set)
         - in case of a generator, the iterable is first converted to a list, then reversed,
         this has a performance impact, and assume a finite generator

         Example:
            <br />
            `q([0, 1, 2, 3, 4]).order()`
            <br />
            `[4, 3, 2, 1, 0]`
        """
        if isinstance(self._items, collections.abc.Generator):
            self._items = reversed(list(self._items))
        else:
            self._items = reversed(self._items)  # type: ignore
        return self

    def slice(self, start: int = 0, stop: Optional[int] = None, step: int = 1) -> 'Query':
        """
        Yields a slice of the iterable.
        Examples:
            <br />
            `q(range(10)).slice(start=1, stop=6, step=2)`
            <br />
            `[1, 3, 5]`

        Args:
            <br />
            start: Optional. The start index of the slice. Defaults to 0.
            <br />
            stop: Optional. The stop index of the slice. Defaults to None.
            <br />
            step: Optional. The step of the slice. Defaults to 1.
        """
        self._items = islice(self._items, start, stop, step)
        return self

    def take(self, n: int, predicate: Optional[Predicate] = None) -> 'Query':
        """
        Yields up to n items that satisfies the predicate (if provided).
        In case the iterable is ordered, the first n items are returned.
        Args:
            <br />
            n: Optional. The number of items to take. Defaults to 1.
            <br />
            predicate: Optional. The predicate to filter the iterable by.
        """
        self.where(predicate)
        self.slice(stop=n)
        return self

    # endregion

    # region Collectors

    def first(self, predicate: Optional[Predicate] = None) -> Any:
        self.where(predicate)
        try:
            return next(self)
        except StopIteration:
            raise NoItemsFoundException()

    def first_or_default(self, default: Any = None, predicate: Optional[Predicate] = None) -> Any:
        self.where(predicate)
        try:
            return next(self)
        except StopIteration:
            return default

    def single(self, predicate: Optional[Predicate] = None) -> Any:
        self.where(predicate)
        try:
            first = next(self)
        except StopIteration:
            raise NoItemsFoundException()

        try:
            next(self)
        except StopIteration:
            return first

        raise MultipleItemsFoundException()

    def single_or_default(self, predicate: Optional[Predicate] = None, default: Any = None) -> Any:
        self.where(predicate)
        try:
            first = next(self)
        except StopIteration:
            return default

        try:
            next(self)
        except StopIteration:
            return first

        raise MultipleItemsFoundException()

    def count(self) -> int:
        """
        Returns the number of elements in the iterable
        :return: The number of the elemtns
        :rtype: int
        """
        # If the iterable is sized, return the length
        if isinstance(self._items, Sized):
            return len(self._items)

        # Otherwise, iterate over the iterable
        return sum(1 for _ in self._items)

    def any(self, predicate: Optional[Predicate] = None) -> bool:
        """
        Returns whether any element in the iterable evaluates to true.
        If a predicate is provided, only elements that satisfy the predicate are considered.
        Args:
            predicate: Optional. The predicate to filter the iterable by.

        Returns:
            True if any element evaluates to true, False otherwise.
        """
        self.where(predicate)
        return any(self._items)

    def to_list(self) -> List:
        return list(self)

    # endregion



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestStreamerDistinct:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_distinct_hasNoItems(self,
                                 iter_type,
                                 iterable,
                                 iterable_list):
        assert list(q(iterable).distinct()) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_exclude_hasSingleItem(self,
                                   iter_type,
                                   iterable,
                                   iterable_list):
        assert list(q(iterable).distinct()) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_exclude_hasMultipleItemsWithoutDups(self,
                                                 iter_type,
                                                 iterable,
                                                 iterable_list):
        assert list(q(iterable).distinct()) == list(iterable_list)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi_dup())
    def test_exclude_hasMultipleItemsWithDups_preserveOrder(self,
                                                            iter_type,
                                                            iterable,
                                                            iterable_list):
        assert list(q(iterable).distinct(preserve_order=True)) == sorted(list(set(iterable_list)))

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi_dup())
    def test_exclude_hasMultipleItemsWithDups_noPreserveOrder(
            self,
            iter_type,
            iterable,
            iterable_list):
        items = set(q(iterable).distinct(preserve_order=False))
        assert items == set(iterable_list)
        assert len(items) == len(set(iterable_list))



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestStreamerSelect:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_select_hasNoItems(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).select(lambda x: x.a)) == []

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_select_hasSingleItem(self, iter_type, iterable):
        assert list(q(iterable).select(lambda x: x.a)) == [0]

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_select_hasMultipleItems(self, iter_type, iterable):
        assert list(q(iterable).select(lambda x: x.a)) == [0, 1, 2, 3, 4]



from unittest import TestCase

from fliq import q


class TestQuery(TestCase):
    def test_query_iterable(self):
        expected = [1, 2, 3]
        actual = [i for i in q(expected)]
        assert expected == actual



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestCollectorCount:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_count_hasNoItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).count() == len(iterable_list)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_get_hasSingleItem(self, iter_type, iterable, iterable_list):
        assert q(iterable).count() == len(iterable_list)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_get_hasMultipleItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).count() == len(iterable_list)



from typing import Generator
from unittest import TestCase

from parameterized import parameterized

from fliq import q
from fliq.tests.fliq_test_utils import MyTestClass, FliqTestUtils
from fliq.tests.timer import Timer


class TestPerformance(TestCase):
    def _query_baseline(self, data: Generator):
        item = next(map(lambda x: x.b, filter(lambda x: x.a == 50, data))) or -1
        return item

    def _query_fliq(self, data: Generator):
        item = (
            q(data)
            .where(lambda x: x.a == 50)
            .select(lambda x: x.b)
            .first_or_default(default=-1)
        )
        return item

    @staticmethod
    def _generate_data(num: int):
        return (MyTestClass(a=i, b=i * 2) for i in range(num))

    @parameterized.expand([
        ("small", 100),
        ("medium", 10_000),
        ("large", 10_000_000)
    ])
    def test_performance_smallDataset(self, name, data):
        self._test_performance(data)

    @FliqTestUtils.retry(attempts=100)
    def _test_performance(self, data, attempt: int):
        expected_item = 100

        with Timer() as baseline_t:
            baseline_item = self._query_baseline(self._generate_data(data))

        with Timer() as fliq_t:
            fliq_item = self._query_fliq(self._generate_data(data))

        self.assertEqual(
            expected_item,
            baseline_item,
            'Baseline query returned incorrect result'
        )
        self.assertEqual(
            expected_item,
            fliq_item,
            'Fliq query returned incorrect result'
        )
        FliqTestUtils.assertSmallerOrCloseTo(
            fliq_t.elapsed,
            baseline_t.elapsed,
            0.01,
            f"Attempt {attempt}"
        )



from dataclasses import dataclass
from unittest import TestCase

from fliq import q

@dataclass
class Item:
    price: int
    rating: float


class TestSanity(TestCase):

    def setUp(self):
        self.items = (Item(i, i % 5) for i in range(100))

    def test_sanity(self):
        price = (
            q(self.items)
            .where(lambda x: x.price > 50)
            .select(lambda x: x.price)
            .first_or_default(default=0)
        )
        assert price == 51



import xml.etree.ElementTree as ET
from pathlib import Path
import sys

# Check if the correct number of arguments is provided
if len(sys.argv) != 2:
    print("Usage: python check_coverage.py <threshold_percentage>")
    sys.exit(1)

# Get the threshold value from the command-line argument (in percentage format)
try:
    threshold_percentage = float(sys.argv[1])
except ValueError:
    print("Threshold must be a valid float value in percentage format")
    sys.exit(1)

# Convert the percentage to a fraction (e.g., 90% to 0.9)
threshold = threshold_percentage / 100.0

# Get the directory of the script (where the script is located)
script_dir = Path(__file__).resolve().parent

# Specify the path to the coverage.xml file in the root of the project
coverage_file_path = script_dir.parent / "coverage.xml"

# Check if the coverage.xml file exists
if not coverage_file_path.exists():
    print("Coverage report not found, failing the workflow")
    sys.exit(1)

tree = ET.parse(coverage_file_path)
root = tree.getroot()

failed_files = []

for class_element in root.findall('.//class[@filename]'):
    filename = class_element.get('filename')
    line_rate_str = class_element.get('line-rate')
    if line_rate_str:
        line_rate = float(line_rate_str)
    else:
        raise AttributeError(f'line-rate attribute not found for file {filename}')

    if line_rate < threshold:
        failed_files.append((filename, line_rate))

if failed_files:
    print(f'Files with coverage below the threshold ({threshold_percentage}):')
    for filename, line_rate in failed_files:
        print(f'- {filename}: {line_rate * 100:.2f}% coverage')
    sys.exit(1)
